@@ -5,7 +5,7 @@ except ImportError:
     import cookielib as cookiejar
 from copy import copy
 from functools import partial
-from contextlib import contextmanager
+from contextlib import contextmanager, nullcontext
 
 from werkzeug.local import LocalStack
 from flask import g, session, get_flashed_messages
@@ -15,9 +15,9 @@ from webtest import (TestApp as BaseTestApp,
                      TestResponse as BaseTestResponse)
 
 try:
-    from flask.ext.sqlalchemy import connection_stack
+    import flask_sqlalchemy
 except ImportError:
-    connection_stack = None
+    flask_sqlalchemy = None
 
 try:
     # Available starting with Flask 0.10
@@ -36,7 +36,7 @@ class SessionScope(object):
     When popped, removes the current session and swap the value of
     :func:`.scopefunc` to the one that was before.
 
-    :param db: :class:`flask.ext.sqlalchemy.SQLAlchemy` instance
+    :param db: :class:`flask_sqlalchemy.SQLAlchemy` instance
     """
 
     def __init__(self, db):
@@ -67,8 +67,22 @@ def get_scopefunc(original_scopefunc=None):
     """
 
     if original_scopefunc is None:
-        assert connection_stack, 'Is Flask-SQLAlchemy installed?'
-        original_scopefunc = connection_stack.__ident_func__
+        assert flask_sqlalchemy, 'Is Flask-SQLAlchemy installed?'
+
+        try:
+            # for flask_sqlalchemy older than 2.2 where the connection_stack
+            # was either the app stack or the request stack
+            original_scopefunc = flask_sqlalchemy.connection_stack.__ident_func__
+        except AttributeError:
+            try:
+                # when flask_sqlalchemy 2.2 or newer, which supports only flask 0.10
+                # or newer, we use app stack
+                from flask import _app_ctx_stack
+                original_scopefunc = _app_ctx_stack.__ident_func__
+            except AttributeError:
+                # newer flask does not expose an __ident_func__, use greenlet directly
+                import greenlet
+                original_scopefunc = greenlet.getcurrent
 
     def scopefunc():
         rv = original_scopefunc()
@@ -172,7 +186,7 @@ class TestApp(BaseTestApp):
     for all requests to the app.
 
     :param app: :class:`flask.Flask` instance
-    :param db: :class:`flask.ext.sqlalchemy.SQLAlchemy` instance
+    :param db: :class:`flask_sqlalchemy.SQLAlchemy` instance
     :param use_session_scopes: if specified, application performs each request
                                within it's own separate session scope
     """
@@ -181,7 +195,7 @@ class TestApp(BaseTestApp):
     def __init__(self, app, db=None, use_session_scopes=False, cookiejar=None,
                  extra_environ=None, *args, **kwargs):
         if use_session_scopes:
-            assert db, ('`db` (instance of `flask.ext.sqlalchemy.SQLAlchemy`) '
+            assert db, ('`db` (instance of `flask_sqlalchemy.SQLAlchemy`) '
                         'must be passed to use session scopes.')
         self.db = db
         self.use_session_scopes = use_session_scopes
@@ -211,8 +225,13 @@ class TestApp(BaseTestApp):
         if self.use_session_scopes:
             scope = SessionScope(self.db)
             scope.push()
+
+        context = nullcontext
+        if self.app.config.get('FLASK_WEBTEST_PUSH_APP_CONTEXT', False):
+            context = self.app.app_context
         try:
-            response = super(TestApp, self).do_request(*args, **kwargs)
+            with context():
+                response = super(TestApp, self).do_request(*args, **kwargs)
         finally:
             if self.use_session_scopes:
                 scope.pop()
